import itertools
from nl4dv.utils import constants, helpers


class TaskGenie:
    def __init__(self, nl4dv_instance):
        self.nl4dv_instance = nl4dv_instance

    def new_task(self, task_name,
                 task_type='implicit',
                 attributes=list(),
                 query_phrase=list(),
                 operator=None,
                 values=None,
                 is_attr_ambiguous=False,
                 is_value_ambiguous=False,
                 value_ambiguity_type=None):

        task = dict()
        task['task'] = task_name
        task['queryPhrase'] = query_phrase
        task['operator'] = operator
        task['values'] = values
        task['matchScore'] = self.nl4dv_instance.match_scores['task'][task_type]
        task['attributes'] = attributes
        task['inferenceType'] = task_type
        task['isAttrAmbiguous'] = is_attr_ambiguous
        task['isValueAmbiguous'] = is_value_ambiguous
        task['meta'] = dict()
        task['meta']['value_ambiguity_type'] = value_ambiguity_type

        return task

    def get_explicit_tasks(self):
        explicit_tasks = set()
        for k in self.nl4dv_instance.extracted_tasks:
            for v in self.nl4dv_instance.extracted_tasks[k]:
                if v["inferenceType"] == 'explicit':
                    explicit_tasks.add(k)

        return explicit_tasks

    def get_implicit_tasks(self):
        implicit_tasks = set()
        for k in self.nl4dv_instance.extracted_tasks:
            for v in self.nl4dv_instance.extracted_tasks[k]:
                if v["inferenceType"] == 'implicit':
                    implicit_tasks.add(k)

        return implicit_tasks

    @staticmethod
    def filter_empty_tasks(task_map):
        tasks_to_delete = set()
        for task_name in task_map:
            if len(task_map[task_name]) == 0:
                tasks_to_delete.add(task_name)

        for task_name in tasks_to_delete:
            del task_name[task_name]
        return task_map

    @staticmethod
    def has_non_filter_explicit_task(task_map):
        for task in task_map:
            if task != "filter" and len(task_map[task]) != 0:
                return True
        return False

    @staticmethod
    def has_non_filter_explicit_task_for_attr_list(task_map, attr_list):
        for task in task_map:
            if task != "filter":
                for task_instance in task_map[task]:
                    if set(task_instance["attributes"]) == set(attr_list):
                        return True

        return False

    def is_datatype_ambiguous(self, attributes, task, values):
        is_datatype_ambiguous = False
        if task == "filter" and len(values) > 0 and helpers.isfloat(values[0]):
            for a in attributes:
                if self.nl4dv_instance.data_genie_instance.data_attribute_map[a]["dataType"] != constants.attribute_types["QUANTITATIVE"]:
                    is_datatype_ambiguous = True
                    break
        elif task in ["derived_value", "find_extremum"]:
            for a in attributes:
                if self.nl4dv_instance.data_genie_instance.data_attribute_map[a]["dataType"] != constants.attribute_types["QUANTITATIVE"]:
                    is_datatype_ambiguous = True
                    break
        elif task in ["trend"]:
            for a in attributes:
                if self.nl4dv_instance.data_genie_instance.data_attribute_map[a]["dataType"] != constants.attribute_types["TEMPORAL"]:
                    is_datatype_ambiguous = True
                    break

        return is_datatype_ambiguous

    def generate_tasks(self,
                       task_name,
                       attributes,
                       query_phrase,
                       operator_phrase,
                       operator,
                       values,
                       inference_type,
                       allow_subset=False):

        task_list = list()
        is_attribute_ambiguous = any(self.nl4dv_instance.extracted_attributes[k]["isAmbiguous"] for k in attributes)
        if is_attribute_ambiguous:
            for idx in range(1, len(attributes) + 1):
                attr_combos = itertools.combinations(attributes, idx)
                for attr_combo in attr_combos:
                    attr_combo = list(attr_combo)

                    if self.nl4dv_instance.attribute_genie_instance.validate_attr_combo(attr_combo=attr_combo,
                                                                                        query_phrase=query_phrase,
                                                                                        allow_subset=allow_subset):
                        continue

                    # Get Attribute datatypes
                    sorted_attr_combo, sorted_attr_type_str = self.nl4dv_instance.attribute_genie_instance.get_attr_datatype_shorthand(attr_combo)

                    # Check if the Value (datatype) is Ambiguous
                    is_value_ambiguous = self.is_datatype_ambiguous(task=task_name, attributes=sorted_attr_combo, values=values)
                    value_ambiguity_type = 'datatype' if is_value_ambiguous else None

                    task_obj = self.new_task(task_name=task_name,
                                             task_type=inference_type,
                                             attributes=sorted_attr_combo,
                                             query_phrase=operator_phrase,
                                             values=values,
                                             operator=operator,
                                             is_attr_ambiguous=is_attribute_ambiguous,
                                             is_value_ambiguous=is_value_ambiguous,
                                             value_ambiguity_type=value_ambiguity_type)

                    if task_obj not in task_list:
                        task_list.append(task_obj)

        else:

            # Check if the Value (datatype) is Ambiguous
            is_value_ambiguous = self.is_datatype_ambiguous(task=task_name, attributes=attributes, values=values)
            value_ambiguity_type = 'datatype' if is_value_ambiguous else None

            task_obj = self.new_task(task_name=task_name,
                                     task_type=inference_type,
                                     attributes=attributes,
                                     query_phrase=operator_phrase,
                                     values=values,
                                     operator=operator,
                                     is_attr_ambiguous=is_attribute_ambiguous,
                                     is_value_ambiguous=is_value_ambiguous,
                                     value_ambiguity_type=value_ambiguity_type)

            if task_obj not in task_list:
                task_list.append(task_obj)

        return task_list

    def extract_explicit_tasks_from_dependencies(self, dependencies):

        # Get encodeable attributes
        encodeable_attributes = self.nl4dv_instance.attribute_genie_instance.get_encodeable_attributes()

        # Initialize the output
        task_map = dict()

        # Infer Tasks from the EXPLICIT TASK UTTERANCES
        if dependencies is not None:

            # Case 0: CORRELATION, and DISTRIBUTION
            k1, k2, operator_phrase = None, None, None
            for dep_index, dep in enumerate(dependencies[0]):

                # (('budget', 'NN'), 'nmod', ('relationship', 'NN'))
                if dep[0][0] in self.nl4dv_instance.keyword_attribute_mapping and dep[2][0] in self.nl4dv_instance.task_keyword_map and dep[1] not in ["case"]:
                    operator_phrase = dep[2][0]
                    if k1:
                        k2 = dep[0][0]
                    else:
                        k1 = dep[0][0]

                # (('relationship', 'NN'), 'nmod', ('budget', 'NN'))
                if dep[2][0] in self.nl4dv_instance.keyword_attribute_mapping and dep[0][0] in self.nl4dv_instance.task_keyword_map and dep[1] not in ["case"]:
                    operator_phrase = dep[0][0]
                    if k1:
                        k2 = dep[2][0]
                    else:
                        k1 = dep[2][0]

                if dep[2][0] in self.nl4dv_instance.keyword_attribute_mapping and dep[0][0] in self.nl4dv_instance.keyword_attribute_mapping and dep[1] == 'conj':
                    k1 = dep[2][0]
                    k2 = dep[0][0]

                # correlate budget and gross for fiction movies
                if dep[2][0] in self.nl4dv_instance.keyword_attribute_mapping and dep[0][0] in self.nl4dv_instance.task_keyword_map and dep[1] == 'conj':
                    k2 = dep[2][0]

                if dep[0][0] in self.nl4dv_instance.keyword_attribute_mapping and dep[2][0] in self.nl4dv_instance.task_keyword_map and dep[1] == 'conj':
                    k2 = dep[0][0]

                if k1 is not None and k2 is not None and k1 != k2 and operator_phrase is not None:

                    # If keywords point to DOMAIN VALUE matches, then do NOT tag them as a correlation task
                    a1 = list(self.nl4dv_instance.keyword_attribute_mapping[k1].keys())[0]
                    a2 = list(self.nl4dv_instance.keyword_attribute_mapping[k2].keys())[0]
                    if self.nl4dv_instance.extracted_attributes[a1]["metric"] == ["attribute_domain_value_match"] or \
                            self.nl4dv_instance.extracted_attributes[a2]["metric"] == ["attribute_domain_value_match"]:
                        continue

                    task = self.nl4dv_instance.task_keyword_map[operator_phrase][0][0]
                    if task not in task_map:
                        task_map[task] = list()

                    # These tasks requires at most 1 Attribute to make sense
                    if task in ["correlation"]:

                        if self.nl4dv_instance.task_keyword_map[operator_phrase][0][0] == task:
                            operator = self.nl4dv_instance.task_keyword_map[operator_phrase][0][1]

                            relevant_attributes = []
                            for k in [k1, k2]:
                                if k in self.nl4dv_instance.keyword_attribute_mapping:
                                    relevant_attributes.extend(list(self.nl4dv_instance.keyword_attribute_mapping[k].keys()))
                                elif self.nl4dv_instance.porter_stemmer_instance.stem(k) in self.nl4dv_instance.keyword_attribute_mapping:
                                    relevant_attributes.extend(list(self.nl4dv_instance.keyword_attribute_mapping[self.nl4dv_instance.porter_stemmer_instance.stem(k)].keys()))

                            _tasks = self.generate_tasks(task_name=task,
                                                         attributes=relevant_attributes,
                                                         operator_phrase=operator_phrase,
                                                         query_phrase=[k1,k2],
                                                         operator=operator,
                                                         values=[],
                                                         inference_type='explicit',
                                                         allow_subset=True)  # IMP, this will be true

                            for _task in _tasks:
                                if _task not in task_map[task]:
                                    task_map[task].append(_task)

                            k1, k2, operator_phrase = None, None, None

            # Case 1: DERIVED_VALUE, FIND_EXTREMUM and TREND. 1 Task Keyword to 1 Attribute Keyword Mapping
            keyword, operator_phrase = None, None
            for dep_index, dep in enumerate(dependencies[0]):

                # OPTIONAL:
                if dep[1] in ['amod', 'nsubj', 'nmod', 'dobj', 'compound', 'dep']:
                    if dep[0][0] in self.nl4dv_instance.keyword_attribute_mapping and dep[2][0] in self.nl4dv_instance.task_keyword_map:
                        operator_phrase = dep[2][0]
                        keyword = dep[0][0]

                    if dep[2][0] in self.nl4dv_instance.keyword_attribute_mapping and dep[0][0] in self.nl4dv_instance.task_keyword_map:
                        operator_phrase = dep[0][0]
                        keyword = dep[2][0]

                if keyword is not None and operator_phrase is not None:
                    task = self.nl4dv_instance.task_keyword_map[operator_phrase][0][0]
                    if task not in task_map:
                        task_map[task] = list()

                    # These tasks requires at most 1 Attribute to make sense
                    if task in ["derived_value", "find_extremum", "trend"]:
                        operator = self.nl4dv_instance.task_keyword_map[operator_phrase][0][1]
                        attributes = list(self.nl4dv_instance.keyword_attribute_mapping[keyword].keys())
                        _tasks = self.generate_tasks(task_name=task,
                                                     attributes=attributes,
                                                     operator_phrase=operator_phrase,
                                                     query_phrase=[keyword],
                                                     operator=operator,
                                                     values=[],
                                                     inference_type='explicit',
                                                     allow_subset=False)

                        for _task in _tasks:
                            if _task not in task_map[task]:
                                task_map[task].append(_task)

                    keyword, operator_phrase = None, None

            # Case 2: Takes care of numeric FILTERS e.g '...less than 50'
            keyword, amount, operator_phrase, has_negation, negation_phrase = None, None, None, False, None
            for dep_index, dep in enumerate(dependencies[0]):

                if dep[0][0] in self.nl4dv_instance.keyword_attribute_mapping and dep[2][1] == 'CD':
                    keyword = dep[0][0]
                    amount = dep[2][0]

                if dep[2][0] in self.nl4dv_instance.keyword_attribute_mapping and dep[0][1] == 'CD':
                    keyword = dep[2][0]
                    amount = dep[0][0]

                if dep[0][0] in self.nl4dv_instance.keyword_attribute_mapping and dep[2][0] in self.nl4dv_instance.task_keyword_map and dep[1] == "neg":
                    has_negation = True
                    negation_phrase = dep[2][0]

                if dep[2][0] in self.nl4dv_instance.keyword_attribute_mapping and dep[0][0] in self.nl4dv_instance.task_keyword_map and dep[1] == "neg":
                    has_negation = True
                    negation_phrase = dep[0][0]

                if dep[0][0] in self.nl4dv_instance.keyword_attribute_mapping and dep[2][0] in self.nl4dv_instance.task_keyword_map:
                    operator_phrase = dep[2][0]
                    keyword = dep[0][0]

                if dep[2][0] in self.nl4dv_instance.keyword_attribute_mapping and dep[0][0] in self.nl4dv_instance.task_keyword_map:
                    operator_phrase = dep[0][0]
                    keyword = dep[2][0]

                if dep[0][1] == 'CD' and dep[2][0] in self.nl4dv_instance.task_keyword_map:
                    operator_phrase = dep[2][0]
                    amount = dep[0][0]

                if dep[2][1] == 'CD' and dep[0][0] in self.nl4dv_instance.task_keyword_map:
                    operator_phrase = dep[0][0]
                    amount = dep[2][0]

                if dep[0][1] == "CD" and dep[2][0] in self.nl4dv_instance.task_keyword_map and dep[1] == "neg":
                    has_negation = True
                    negation_phrase = dep[2][0]

                if dep[2][1] == "CD" and dep[0][0] in self.nl4dv_instance.task_keyword_map and dep[1] == "neg":
                    has_negation = True
                    negation_phrase = dep[0][0]

                if keyword is not None and amount is not None and operator_phrase is not None:
                    # print(keyword, amount, operator_phrase)
                    if self.nl4dv_instance.task_keyword_map[operator_phrase][0][0] == "filter" and operator_phrase != "between":
                        operator = self.nl4dv_instance.task_keyword_map[operator_phrase][0][1]
                        if has_negation:
                            if operator == "GT":
                                operator = "LT"
                            elif operator == "LT":
                                operator = "GT"
                            operator_phrase = negation_phrase + " " + operator_phrase

                        task = 'filter'
                        if task not in task_map:
                            task_map[task] = list()

                        _tasks = self.generate_tasks(task_name=task,
                                                     attributes=list(self.nl4dv_instance.keyword_attribute_mapping[keyword].keys()),
                                                     operator_phrase=operator_phrase,
                                                     query_phrase=[keyword],
                                                     operator=operator,
                                                     values=[float(amount)],
                                                     inference_type='explicit',
                                                     allow_subset=False)

                        for _task in _tasks:
                            if _task not in task_map[task]:
                                task_map[task].append(_task)

                        keyword, amount, operator_phrase, has_negation, negation_phrase = None, None, None, False, None

            # Case 3: Takes care of numeric FILTERS e.g '...between 5 and 10'
            keyword, from_amount, to_amount, operator_phrase, has_negation, negation_phrase = None, None, None, None, False, None
            for dep_index, dep in enumerate(dependencies[0]):

                if dep[0][1] == 'CD' and dep[1] == 'conj' and dep[2][1] == 'CD':
                    from_amount = dep[0][0]
                    to_amount = dep[2][0]

                if dep[0][1] == 'CD' and dep[2][0] in self.nl4dv_instance.keyword_attribute_mapping:
                    keyword = dep[2][0]

                if dep[2][1] == 'CD' and dep[0][0] in self.nl4dv_instance.keyword_attribute_mapping:
                    keyword = dep[0][0]

                if dep[0][1] == "CD" and dep[2][0] in self.nl4dv_instance.task_keyword_map and dep[1] == "neg":
                    has_negation = True
                    negation_phrase = dep[2][0]

                if dep[2][1] == "CD" and dep[0][0] in self.nl4dv_instance.task_keyword_map and dep[1] == "neg":
                    has_negation = True
                    negation_phrase = dep[0][0]

                if dep[0][1] == 'CD' and dep[2][0] in self.nl4dv_instance.task_keyword_map:
                    operator_phrase = dep[2][0]

                if dep[2][1] == 'CD' and dep[0][0] in self.nl4dv_instance.task_keyword_map:
                    operator_phrase = dep[0][0]

                if keyword is not None and from_amount is not None and to_amount is not None and operator_phrase is not None:

                    if self.nl4dv_instance.task_keyword_map[operator_phrase][0][0] == "filter" and operator_phrase == "between":
                        operator = self.nl4dv_instance.task_keyword_map[operator_phrase][0][1]
                        if has_negation:
                            # NOT RANGE
                            operator = "NOT" + " " + operator
                            operator_phrase = negation_phrase + " " + operator_phrase

                        task = 'filter'
                        if task not in task_map:
                            task_map[task] = list()

                        _tasks = self.generate_tasks(task_name=task,
                                                     attributes=list(self.nl4dv_instance.keyword_attribute_mapping[keyword].keys()),
                                                     operator_phrase=operator_phrase,
                                                     query_phrase=[keyword],
                                                     operator=operator,
                                                     values=[float(from_amount), float(to_amount)],
                                                     inference_type='explicit',
                                                     allow_subset=False)

                        for _task in _tasks:
                            if _task not in task_map[task]:
                                task_map[task].append(_task)

                        keyword, amount, operator_phrase, has_negation, negation_phrase = None, None, None, False, None

            # Case 4: DISTRIBUTION single task keyword
            for operator_phrase in self.nl4dv_instance.task_keyword_map:
                if self.nl4dv_instance.task_keyword_map[operator_phrase][0][0] == "distribution":
                    if operator_phrase in self.nl4dv_instance.query_processed:
                        # It exists, create distribution task
                        task = 'distribution'
                        if task not in task_map:
                            task_map[task] = list()

                        _tasks = self.generate_tasks(task_name=task,
                                                     attributes=encodeable_attributes,
                                                     operator_phrase=operator_phrase,
                                                     query_phrase=[list(self.nl4dv_instance.attribute_keyword_mapping[a].keys())[0] for a in encodeable_attributes],
                                                     operator=None,
                                                     values=[],
                                                     inference_type='explicit',
                                                     allow_subset=False)

                        for _task in _tasks:
                            if _task not in task_map[task]:
                                task_map[task].append(_task)

        return task_map

    def extract_explicit_tasks_from_domain_value(self, task_map):

        # Case 3: Takes care of IMPLICIT categorical filters (from the domain)
        for attr, attr_obj in self.nl4dv_instance.extracted_attributes.items():
            if 'attribute_domain_value_match' in attr_obj['metric']:
                task = 'filter'
                attributes = [attr]
                is_value_ambiguous = False
                value_ambiguity_type = None
                values = list()

                if task not in task_map:
                    task_map[task] = list()

                for k in attr_obj["meta"]["ambiguity"]:
                    values.extend(attr_obj["meta"]["ambiguity"][k])
                    if len(attr_obj["meta"]["ambiguity"][k]) > 1:
                        is_value_ambiguous = True
                        value_ambiguity_type = 'domain_value'

                task_obj = self.new_task(task_name=task,
                                         task_type='explicit',
                                         attributes=attributes,
                                         query_phrase=attr_obj["queryPhrase"],
                                         values=values,
                                         operator="IN",
                                         is_attr_ambiguous=attr_obj["isAmbiguous"],
                                         is_value_ambiguous=is_value_ambiguous,
                                         value_ambiguity_type=value_ambiguity_type)

                if task_obj not in task_map[task]:
                    task_map[task].append(task_obj)

        return task_map

    def extract_implicit_tasks_from_attributes(self, task_map, attribute_list):
        # IMPLICITLY infer tasks based on Attributes (Counts and Datatypes)
        # IF UNABLE to detect from Visualizations and/or Explicit utterances in the query
        # E.g. trend (from temporal)
        # E.g. correlation (from two quantitative), distribution (from one quantitative, one nominal/ordinal)

        if not self.has_non_filter_explicit_task(task_map):
            for i in range(1, len(attribute_list) + 1):
                combinations = itertools.combinations(attribute_list, i)
                for combination in combinations:
                    combo = list(combination)
                    if self.nl4dv_instance.attribute_genie_instance.validate_attr_combo(attr_combo=combo, query_phrase=[], allow_subset=False):
                        continue

                    # Get attribute datatypes
                    sorted_attr_combo, sorted_attr_datatype_combo_str = self.nl4dv_instance.attribute_genie_instance.get_attr_datatype_shorthand(combo)
                    if self.has_non_filter_explicit_task_for_attr_list(task_map, sorted_attr_combo):
                        continue

                    # Keeping it outside of the if elif
                    if 'T' in sorted_attr_datatype_combo_str and not sorted_attr_datatype_combo_str in ["QQT"]:
                        # Add TREND task
                        task = 'trend'
                        attributes = sorted_attr_combo

                        if task not in task_map:
                            task_map[task] = list()

                        task_obj = self.new_task(task_name=task,
                                                 task_type='implicit',
                                                 attributes=attributes,
                                                 is_attr_ambiguous=any(self.nl4dv_instance.extracted_attributes[x]["isAmbiguous"] for x in attributes))
                        if task_obj not in task_map[task]:
                            task_map[task].append(task_obj)

                    if sorted_attr_datatype_combo_str in ["QQ","QQN","QQO","QQQ","QQT"]:
                        # Add CORRELATION task
                        task = 'correlation'
                        attributes = sorted_attr_combo[0:2]

                        if task not in task_map:
                            task_map[task] = list()

                        task_obj = self.new_task(task_name=task,
                                                 task_type='implicit',
                                                 attributes=attributes,
                                                 is_attr_ambiguous=any(self.nl4dv_instance.extracted_attributes[x]["isAmbiguous"] for x in attributes))

                        if task_obj not in task_map[task]:
                            task_map[task].append(task_obj)

                    if sorted_attr_datatype_combo_str in ["QN","QO","QNN","QNO","QOO"]:
                        # Add Derived Value task
                        task = 'derived_value'
                        attributes = [sorted_attr_combo[0]]

                        if task not in task_map:
                            task_map[task] = list()

                        task_obj = self.new_task(task_name=task,
                                                 task_type='implicit',
                                                 attributes=attributes,
                                                 operator='AVG',
                                                 is_attr_ambiguous=any(self.nl4dv_instance.extracted_attributes[x]["isAmbiguous"] for x in attributes))

                        if task_obj not in task_map[task]:
                            task_map[task].append(task_obj)

                    if sorted_attr_datatype_combo_str in ["Q","N","O","NN","NO","OO"]:
                        # Add Distribution task
                        task = 'distribution'
                        attributes = sorted_attr_combo

                        if task not in task_map:
                            task_map[task] = list()

                        task_obj = self.new_task(task_name=task,
                                                 task_type='implicit',
                                                 attributes=attributes,
                                                 is_attr_ambiguous=any(self.nl4dv_instance.extracted_attributes[x]["isAmbiguous"] for x in attributes))

                        if task_obj not in task_map[task]:
                            task_map[task].append(task_obj)

        return task_map



import re
from nltk.corpus import wordnet as wn
from itertools import product
import math
from collections import Counter, MutableMapping
from fuzzywuzzy import fuzz
import Levenshtein
# from dateparser import parse
from dateutil.parser import parse
from datetime import date, datetime
import copy
from nl4dv.utils import constants

WORD = re.compile(r'\w+')


# Includes Repetition
def get_ngrams(input, n):
    input = input.split(' ')
    output = []
    for i in range(len(input)-n+1):
        output.append(input[i:i+n])
    return output


# Removes Duplicate N-Grams
def get_ngrams_without_duplicate(input, n):
    input = input.split(' ')
    output = {}
    for i in range(len(input)-n+1):
        g = ' '.join(input[i:i+n])
        output.setdefault(g, 0)
        output[g] += + 1
        return output


# ToDo:- Explore alternatives to WordNet similarity
def synonymity_score(word_x, word_y):
    # type: (str, str) -> int
    """
    Helper method to find similarity between two words

    """
    sem_1 = wn.synsets(word_x)
    sem_2 = wn.synsets(word_y)
    max_score = 0
    for i, j in list(product(*[sem_1, sem_2])):
        score = i.wup_similarity(j)
        max_score = score if (score is not None and max_score < score) else max_score

    return max_score * 100


def set_default(obj):
    """
    Return a serialized version of set  which is a list.

    """
    if isinstance(obj, (datetime, date)):
        return obj.isoformat()

    if isinstance(obj, set):
        return list(obj)

    raise TypeError ("Type %s not serializable" % type(obj))


def cond_print(string_to_print, debug=True):
    if debug:
        print(string_to_print)


def normalize(vis_list, attr, tasks):
    max_score_by_attr = len(attr)
    max_score_by_tasks = len(tasks)
    max_score_by_vis = 1
    for vis in vis_list:
        vis["normalized_score"] = vis["score"] / (max_score_by_attr + max_score_by_tasks + max_score_by_vis)
    return vis_list


def vectorize(text):
    words = WORD.findall(text)
    return Counter(words)


def cosine_similarity(vec1, vec2):
    intersection = set(vec1.keys()) & set(vec2.keys())
    numerator = sum([vec1[x] * vec2[x] for x in intersection])

    sum1 = sum([vec1[x]**2 for x in vec1.keys()])
    sum2 = sum([vec2[x]**2 for x in vec2.keys()])
    denominator = math.sqrt(sum1) * math.sqrt(sum2)

    if not denominator:
        return 0.0
    else:
        score = float(numerator) * 100 / denominator
        return score


def compute_similarity(str1, str2, type):
    if type == 'cosine_similarity':
        return cosine_similarity(vectorize(str1), vectorize(str2.lower()))

    elif type == 'token_similarity':
        return fuzzy_token_similarity(str1, str2.lower())

    elif type == 'partial_ratio':
        return fuzzy_partial_ratio(str1, str2.lower())

    elif type == 'ratio_similarity':
        return fuzzy_ratio_similarity(str1, str2.lower())

    elif type == 'levenshtein_distance':
        return levenshtein_distance(str1, str2.lower())


def fuzzy_token_similarity(str1, str2):
    return fuzz.token_set_ratio(str1, str2)


def fuzzy_ratio_similarity(str1, str2):
    return fuzz.ratio(str1, str2)


def fuzzy_partial_ratio(str1, str2):
    return fuzz.partial_ratio(str1, str2)


def levenshtein_distance(str1, str2):
    dist = Levenshtein.distance(str1, str2)
    longer_str = str1 if len(str1) > len(str2) else str2
    score = float(dist) * 100 / len(longer_str)
    return score


def isfloat(datum):
    try:
        if datum == '' or str(datum).isspace():
            return False
        float(datum)
    except AttributeError:
        return False
    except ValueError:
        return False
    except OverflowError:
        return False
    return True


def isint(datum):
    try:
        if datum == '' or str(datum).isspace():
            return False
        a = float(datum)
        b = int(a)
    except AttributeError:
        return False
    except ValueError:
        return False
    except OverflowError:
        return False
    return a == b


# # Dateparser settings
# dateparser_settings = dict()
# dateparser_settings['PREFER_DAY_OF_MONTH'] = 'first'
# dateparser_settings['RELATIVE_BASE'] = datetime(2020, 1, 1)
def isdate(datum):
    try:
        if datum == '' or str(datum).isspace():
            return False, None
        return True, parse(datum, fuzzy=False)
        # parsed_obj = parse(datum, settings=constants.dateparser_settings)
        # return parsed_obj is not None, parsed_obj
    except AttributeError:
        return False, None
    except ValueError:
        return False, None
    except OverflowError:
        return False, None


# Trim the output LIST based on the debug parameter
def delete_keys_from_list(array, keys):
    keys_set = set(keys)
    modified_array = copy.deepcopy(array)
    for index, value in enumerate(array):
        if isinstance(value, MutableMapping):
            modified_array[index] = delete_keys_from_dict(value, keys_set)
        else:
            modified_array[index] = value
    return modified_array


# Trim the output DICTIONARY based on the debug parameter
def delete_keys_from_dict(dictionary, keys):
    keys_set = set(keys)
    modified_dict = {}
    for key, value in dictionary.items():
        if key not in keys_set:
            if isinstance(value, MutableMapping):
                modified_dict[key] = delete_keys_from_dict(value, keys_set)
            elif isinstance(value, list):
                modified_dict[key] = delete_keys_from_list(value, keys_set)
            else:
                modified_dict[key] = value  # or copy.deepcopy(value) if a copy is desired for non-dicts.
    return modified_dict


# Use to determine if two lists have any common elements, if any.
def common_member(a, b):
    a_set = set(a)
    b_set = set(b)
    if (a_set & b_set):
        return True
    else:
        return False



from nl4dv import *
from nl4dv.utils import constants
import pytest
import os


class TestDataGenieClass:

    dataset_attribute_type_mapping = {
        'euro': {
            "Foot": constants.attribute_types["NOMINAL"],
            "Name": constants.attribute_types["NOMINAL"],
            "Position": constants.attribute_types["NOMINAL"],
            "Club": constants.attribute_types["NOMINAL"],
            "Country": constants.attribute_types["NOMINAL"],
            "Age": constants.attribute_types["QUANTITATIVE"],
            "Salary": constants.attribute_types["QUANTITATIVE"],
            "Goals": constants.attribute_types["QUANTITATIVE"]
        },
        'colleges': {
            "ACT Median": constants.attribute_types["QUANTITATIVE"],
            "Admission Rate": constants.attribute_types["QUANTITATIVE"],
            "Average Cost": constants.attribute_types["QUANTITATIVE"],
            "Average Faculty Salary": constants.attribute_types["QUANTITATIVE"],
            "Control": constants.attribute_types["NOMINAL"],
            "Expenditure": constants.attribute_types["QUANTITATIVE"],
            "Locale": constants.attribute_types["NOMINAL"],
            "Median Debt": constants.attribute_types["QUANTITATIVE"],
            "Median Earnings": constants.attribute_types["QUANTITATIVE"],
            "Median Family Income": constants.attribute_types["QUANTITATIVE"],
            "Name": constants.attribute_types["NOMINAL"],
            "Population": constants.attribute_types["QUANTITATIVE"],
            "Region": constants.attribute_types["NOMINAL"],
            "SAT Average": constants.attribute_types["QUANTITATIVE"]
        },
        'movies-w-year': {
            "Content Rating": constants.attribute_types["NOMINAL"],
            "Creative Type": constants.attribute_types["NOMINAL"],
            "Genre": constants.attribute_types["NOMINAL"],
            "IMDB Rating": constants.attribute_types["QUANTITATIVE"],
            "Production Budget": constants.attribute_types["QUANTITATIVE"],
            "Release Year": constants.attribute_types["QUANTITATIVE"],
            "Rotten Tomatoes Rating": constants.attribute_types["QUANTITATIVE"],
            "Running Time": constants.attribute_types["QUANTITATIVE"],
            "Title": constants.attribute_types["NOMINAL"],
            "Worldwide Gross": constants.attribute_types["QUANTITATIVE"]
        },
        'housing': {
            "Basement Area": constants.attribute_types["QUANTITATIVE"],
            "Central Air": constants.attribute_types["NOMINAL"],
            "Fence Type": constants.attribute_types["NOMINAL"],
            "Fireplaces": constants.attribute_types["ORDINAL"],
            "Foundation Type": constants.attribute_types["NOMINAL"],
            "Garage Type": constants.attribute_types["NOMINAL"],
            "Heating Type": constants.attribute_types["NOMINAL"],
            "Home Type": constants.attribute_types["NOMINAL"],
            "Lot Area": constants.attribute_types["QUANTITATIVE"],
            "Lot Config": constants.attribute_types["NOMINAL"],
            "Price": constants.attribute_types["QUANTITATIVE"],
            "Roof Style": constants.attribute_types["NOMINAL"],
            "Rooms": constants.attribute_types["ORDINAL"],
            "Satisfaction": constants.attribute_types["ORDINAL"],
            "Year": constants.attribute_types["ORDINAL"]
        },
        'olympic-medals': {
            "Age": constants.attribute_types["QUANTITATIVE"],
            "Bronze Medal": constants.attribute_types["QUANTITATIVE"],
            "Country": constants.attribute_types["NOMINAL"],
            "Gold Medal": constants.attribute_types["QUANTITATIVE"],
            "Name": constants.attribute_types["NOMINAL"],
            "Sex": constants.attribute_types["NOMINAL"],
            "Silver Medal": constants.attribute_types["QUANTITATIVE"],
            "Sport": constants.attribute_types["NOMINAL"],
            "Total Medal": constants.attribute_types["QUANTITATIVE"],
            "Year": constants.attribute_types["QUANTITATIVE"]
        },
        'cars-w-year': {
            "Acceleration": constants.attribute_types["QUANTITATIVE"],
            "Cylinders": constants.attribute_types["ORDINAL"],
            "Displacement": constants.attribute_types["QUANTITATIVE"],
            "Horsepower": constants.attribute_types["QUANTITATIVE"],
            "MPG": constants.attribute_types["QUANTITATIVE"],
            "Model": constants.attribute_types["NOMINAL"],
            "Origin": constants.attribute_types["NOMINAL"],
            "Weight": constants.attribute_types["QUANTITATIVE"],
            "Year": constants.attribute_types["QUANTITATIVE"]
        }
    }

    # Dependency Parser OPTIONS
    dependency_parser_config = {
        "corenlp": {'name': 'corenlp', 'model': os.path.join("examples","assets","jars","stanford-english-corenlp-2018-10-05-models.jar"),'parser': os.path.join("examples","assets","jars","stanford-parser.jar")},
        "corenlp-server": {'name': 'corenlp-server', 'url': 'http://localhost:9000'},
        "spacy": {'name': 'spacy','model': 'en_core_web_sm'}
    }

    @staticmethod
    @pytest.fixture(scope='class')
    def set_dependency_parser(request):
        # Base NL4DV instance
        return request.param

    @staticmethod
    @pytest.fixture(scope='class')
    def init_nl4dv(request):

        # Access request param and assign to dataset
        dataset = request.param

        # Base NL4DV instance
        nl4dv_instance = NL4DV(verbose=False)
        nl4dv_instance.set_data(data_url=os.path.join("examples","assets","data",dataset + ".csv"))
        nl4dv_instance.set_alias_map(alias_url=os.path.join("examples","assets","aliases",dataset + ".json"))
        return dataset, nl4dv_instance

    @pytest.mark.parametrize("init_nl4dv", list(dataset_attribute_type_mapping.keys()), indirect=True, scope="class")
    @pytest.mark.parametrize("set_dependency_parser", list(dependency_parser_config.keys()), indirect=True, scope="class")
    def test_attribute_datatypes(self, init_nl4dv, set_dependency_parser):
        dataset = init_nl4dv[0]
        nl4dv_instance = init_nl4dv[1]
        nl4dv_instance.set_dependency_parser(config=TestDataGenieClass.dependency_parser_config[set_dependency_parser])
        data_summary = nl4dv_instance.get_metadata()
        errors = []
        for attr in data_summary:
            if data_summary[attr]["dataType"] != TestDataGenieClass.dataset_attribute_type_mapping[dataset][attr]:
                errors.append(1)

        assert len(errors) == 0



# Map of "visualizations" that can be explicitly requested and their corresponding keywords to look for in the query.
# allow ' ' (space) in keywords? eg. bar chart / scatter plot ,etc ?
vis_keyword_map = {
    'barchart': {'bar chart', 'barchart'},
    'scatterplot': {'scatter plot', 'scatterplot', 'scatter'},
    'boxplot': {'box plot', 'boxplot'},
    'linechart': {'time series', 'timeseries', 'linechart', 'line chart'},
    'areachart': {'area chart', 'areachart'},
    'stripplot': {'tick plot', 'tickplot', 'strip plot', 'stripplot'},
    'piechart': {'pie chart', 'piechart'},
    'donutchart': {'donut chart', 'donutchart'},
    'histogram': {'histogram'}
}

vis_to_task_mapping = {
    'barchart': ['distribution','derived_value'],
    'scatterplot': ['correlation'],
    'boxplot': ['distribution'],
    'linechart': ['trend'],
    'areachart': ['trend'],
    'stripplot': ['distribution'],
    'piechart': ['distribution','derived_value'],
    'donutchart': ['distribution','derived_value'],
    'histogram': ['distribution']
}

# Map of "tasks" that can be performed and their corresponding keywords to look for in the query.
task_keyword_map = {

    # -----------------------------------------
    # Find Extremum
    # -----------------------------------------
    "rank": [("find_extremum", "MIN")],
    "sort": [("find_extremum", "MIN")],
    "max": [("find_extremum", "MAX")],
    "maximum": [("find_extremum", "MAX")],
    "highest": [("find_extremum", "MAX")],
    "greatest": [("find_extremum", "MAX")],
    "largest": [("find_extremum", "MAX")],
    "biggest": [("find_extremum", "MAX")],
    "most": [("find_extremum", "MAX")],
    "min": [("find_extremum", "MIN")],
    "minimum": [("find_extremum", "MIN")],
    "smallest": [("find_extremum", "MIN")],
    "lowest": [("find_extremum", "MIN")],
    "least": [("find_extremum", "MIN")],  # Todo: dep parser can't detect
    "heaviest": [("find_extremum", "MAX")],
    "lightest": [("find_extremum", "MIN")],
    "best": [("find_extremum", "MAX")],
    "worst": [("find_extremum", "MIN")],
    # ADd SORT
    # -----------------------------------------

    # -----------------------------------------
    # Filter
    # -----------------------------------------
    "after": [("filter", "GT")],  # Todo: dep parser can't detect
    "before": [("filter", "LT")],  # Todo: dep parser can't detect
    "more": [("filter", "GT")],
    "high": [("filter", "GT")],
    "over": [("filter", "GT")],
    "higher": [("filter", "GT")],
    "greater": [("filter", "GT")],
    "larger": [("filter", "GT")],
    "bigger": [("filter", "GT")],
    "under": [("filter", "LT")],
    "less": [("filter", "LT")],
    "lower": [("filter", "LT")],
    "lesser": [("filter", "LT")],
    "smaller": [("filter", "LT")],
    "between": [("filter", "RANGE")],
    # -----------------------------------------

    # -----------------------------------------
    # Compute Derived Value
    # -----------------------------------------
    "median": [("derived_value", "MEDIAN")],
    "average": [("derived_value", "AVG")],
    "mean": [("derived_value", "AVG")],
    "sum": [("derived_value", "SUM")],
    "total": [("derived_value", "SUM")],
    # -----------------------------------------

    # -----------------------------------------
    # Characterize Distribution
    # -----------------------------------------
    "distribution": [("distribution", None)],
    "range": [("distribution", None)],
    "extent": [("distribution", None)],
    # -----------------------------------------

    # -----------------------------------------
    # Outliers / Anomalies
    # -----------------------------------------
    "outlier": [("outlier",)],
    "exception": [("outlier",)],
    # -----------------------------------------

    # -----------------------------------------
    # Correlation
    # -----------------------------------------
    "correlation": [("correlation", None)],
    "correlate": [("correlation", None)],
    "relation": [("correlation", None)],
    "relationship": [("correlation", None)],
    "relate": [("correlation", None)],
    # -----------------------------------------
    # Negation
    # -----------------------------------------
    "not": [("negation", None)],

    # -----------------------------------------
    # Equality
    # -----------------------------------------
    "equal": [("filter", "EQ")],

    # -----------------------------------------
    # Trend
    # -----------------------------------------
    "trend": [("trend", None)]

    # ToDo:- Detect Negations using dependency parsing and/or algorithms !

}

# scores given to attribute/ task/ vis matches of different types
match_scores = {
    'attribute': {
        'attribute_similarity_match': 1,
        'attribute_alias_similarity_match': 0.75,
        'attribute_synonym_match': 0.5,
        'attribute_domain_value_match': 0.5,
    },
    'task': {
        'explicit': 1,
        'implicit': 0.5,
    },
    'vis': {
        'explicit': 1
    }
}

# JSON has an abbreviated version (GT) but Vega-Lite requires symbols (>)
operator_symbol_mapping = {
    "NOT RANGE": "not between",
    "GT": ">",
    "LT": "<",
    "EQ": "==",
    "SUM": "sum",
    "MEDIAN": "median",
    "AVG": "mean",
    "IN": "in",
    "RANGE": "between",
    "MAX": "max",
    "MIN": "min"
}

# keep them high, so single letter typos in shorter words are not caught. eg. with > width
match_thresholds = {
    'synonymity': 95,
    'string_similarity': 85
}

# These keys are deleted from final output since they are NOT important/critical. They are retained when the `debug` flag is true.
keys_to_delete_in_output = ['matchScore',
                            'meta',
                            'debug',
                            'score',
                            'status',
                            'scoreObj',
                            'encode',
                            'isLabel',
                            'metric',
                            'isAttrAmbiguous',
                            'isValueAmbiguous',
                            'confidenceObj']

attribute_types = {
    'QUANTITATIVE': 'Q',
    'NOMINAL': 'N',
    'TEMPORAL': 'T',
    'ORDINAL': 'O',
}

# NL4DV datatypes to VL datatypes mapping
vl_attribute_types = {
    'Q': 'quantitative',
    'N': 'nominal',
    'T': 'temporal',
    'O': 'ordinal',
}



from datetime import datetime
from itertools import product
import dateparser
from dateutil.parser import parse
from gensim.models import Word2Vec, KeyedVectors
from nltk.corpus import wordnet as wn

def word2vec_stuff():
    # Use the BROWN corpus from NLTK
    # model = Word2Vec(brown.sents())
    # model.save('examples/assets/jars/brown.embedding')
    model = Word2Vec.load('examples/assets/jars/brown.embedding')
    # vocabulary = model.wv.vocab
    print(model.similarity('expensive', 'price'))
    # print(model.most_similar(positive=['price'], topn = 3))

    # Use Google News corpus: https://code.google.com/archive/p/word2vec/
    # model = KeyedVectors.load_word2vec_format('examples/assets/jars/GoogleNews-vectors-negative300.bin.gz', binary=True)
    # model.save('examples/assets/jars/googlenews.embedding')
    model = KeyedVectors.load('examples/assets/jars/googlenews.embedding')
    print(model.similarity('expensive', 'price'))

def isdateParser(datum):
    settings = dict()
    settings['PREFER_DAY_OF_MONTH'] = 'first'
    settings['RELATIVE_BASE'] = datetime(2020, 1, 1)
    try:
        if datum == '' or str(datum).isspace():
            return False
        print('dateparser: ', dateparser.parse(datum, settings=settings), '  dateutil.parser: ', parse(datum, fuzzy=False))
    except AttributeError:
        return False
    except ValueError:
        return False
    except OverflowError:
        return False
    return True


def dateparsing():
    isdateParser("12/05/18")
    isdateParser("12-05-18")
    isdateParser("December 2018")
    isdateParser("Dec 2018")
    isdateParser("Dec 5 2018")
    isdateParser("December 5 2018")
    isdateParser("2018")
    isdateParser("5 Dec 2018")
    isdateParser("8 June")
    isdateParser("June 8")


def synonymity_score(word_x, word_y):
    # type: (str, str) -> int
    """
    Helper method to find similarity between two words

    """
    sem_1 = wn.synsets(word_x)
    sem_2 = wn.synsets(word_y)
    max_score = 0
    for i, j in list(product(*[sem_1, sem_2])):
        score = i.wup_similarity(j)
        max_score = score if (score is not None and max_score < score) else max_score

    return max_score * 100


# word2vec_stuff()
dateparsing()
# print (synonymity_score("date","year"))



from nl4dv.utils import constants, helpers
import re
import nltk
from nltk import word_tokenize
from si_prefix import si_parse
from nltk.corpus import stopwords


class QueryGenie:
    def __init__(self, nl4dv_instance):
        self.nl4dv_instance = nl4dv_instance

        # Stopwords
        self.stopwords_set = set(stopwords.words('english'))

    def process_query(self, query):

        # Try to infer and convert numerical shorthands to machine understandable format, e.g. 100M = 100*1,000,000
        parsed_tokens = []
        for token in word_tokenize(query):
            try:
                parsed_tokens.append(str(int(si_parse(token))))
            except Exception as e:
                parsed_tokens.append(token)

        # Clean sentence of non-alphanumerical characters
        query_processed = re.sub(r'[^A-Za-z0-9]+', ' ', ' '.join(parsed_tokens))

        # Convert to lowercase
        query_lower = query_processed.lower()

        return query_lower

    def clean_query_and_get_query_tokens(self, query, reserve_words, ignore_words):

        # Set the stopwords from reserve words and ignore words
        self.stopwords_set = self.stopwords_set.difference(set(reserve_words)).union(set(ignore_words))

        # Create token set and filter out standard stopwords
        query_tokens = list(filter(lambda token: token not in self.stopwords_set, word_tokenize(query)))

        return query_tokens

    # 1-time generation of query N-GRAMS
    def get_query_ngrams(self, query):
        ngrams = dict()
        for i in range(len(query.split()), 0, -1):
            for ngram in helpers.get_ngrams(query, i):
                ngram_str = ((' '.join(map(str, ngram))).rstrip()).lower()
                ngrams[ngram_str] = dict()
                ngrams[ngram_str]['raw'] = ngram
                ngrams[ngram_str]['lower'] = ngram_str
                ngrams[ngram_str]['stemmed_lower'] = ' '.join(self.nl4dv_instance.porter_stemmer_instance.stem(t) for t in nltk.word_tokenize(ngram_str))

        return ngrams

    # Create a parts of speech map from the Query Tokens
    def get_query_pos_map(self, query_tokens):
        pos_map = dict()
        pos_list = nltk.pos_tag(query_tokens)
        for pos in pos_list:
            pos_map[pos[0]] = pos[1]

    # Create  a Dependency Tree from the query
    def create_dependency_tree(self, query):

        if self.nl4dv_instance.dependency_parser == "spacy":
            doc = self.nl4dv_instance.dependency_parser_instance(query)
            dependencies = [[]]
            for token in doc:
                dependency = [(token.head.text, token.head.tag_), token.dep_, (token.text, token.tag_)]
                dependencies[0].append(dependency)
                # print(
                #     "text:", token.text, " | "
                #     "pos:", token.pos_,  " | "
                #     "tag:", token.tag_,  " | "
                #     "dependency:", token.dep_,  " | "
                #     "htext:", token.head.text, " | "
                #     "hpos:", token.head.pos_,  " | "
                #     "htag:", token.head.tag_,  " | "
                #     "hdependency:", token.head.dep_,  " | "
                #     "subtree:", [s for s in token.subtree],
                #     "treechildren:", [child for child in token.children]
                #     # "lemma:", token.lemma_,  " | "
                #     # "shape:", token.shape_,  " | "
                #     # "isalphabet:", token.is_alpha, " | "
                #     # "isstopword:", token.is_stop,  " | "
                # )
                # print("{2}({3}-{6}, {0}-{5})".format(token.text, token.tag_, token.dep_, token.head.text, token.head.tag_, token.i+1, token.head.i+1))

            # print("\nSpacy Noun Chunks:")
            # print("------------------------------")
            # for chunk in doc.noun_chunks:
            #     print(
            #        "text:", chunk.text,
            #         "rtext:", chunk.root.text,
            #         "rdep:", chunk.root.dep_,
            #         "htext:", chunk.root.head.text
            #     )
            #
            # print("\nSpacy Named Entity Recognition:")
            # print("------------------------------")
            # for ent in doc.ents:
            #     print(ent.text, ent.start_char, ent.end_char, ent.label_)

            return dependencies

        elif self.nl4dv_instance.dependency_parser == "corenlp":
            dependencies = [list(parse.triples()) for parse in self.nl4dv_instance.dependency_parser_instance.raw_parse(query)]

            # Encode every string in tree to utf8 so string matching will work
            for dependency in dependencies[0]:
                dependency[0][0].encode('utf-8')
                dependency[0][1].encode('utf-8')
                dependency[1].encode('utf-8')
                dependency[2][0].encode('utf-8')
                dependency[2][1].encode('utf-8')

            return dependencies

        elif self.nl4dv_instance.dependency_parser == "corenlp-server":
            dependencies = [list(parse.triples()) for parse in self.nl4dv_instance.dependency_parser_instance.raw_parse(query)]

            # Encode every string in tree to utf8 so string matching will work
            for dependency in dependencies[0]:
                dependency[0][0].encode('utf-8')
                dependency[0][1].encode('utf-8')
                dependency[1].encode('utf-8')
                dependency[2][0].encode('utf-8')
                dependency[2][1].encode('utf-8')

            return dependencies



import itertools
from nl4dv.vlgenie import VLGenie
from nl4dv.visgenie.vis_recos import vis_design_combos
from nl4dv.utils import constants
import copy


class VisGenie:

    def __init__(self, nl4dv_instance):
        self.nl4dv_instance = nl4dv_instance

    def extract_vis_type(self, query_ngrams):
        for ngram in query_ngrams:
            for vis_type, vis_keywords in self.nl4dv_instance.vis_keyword_map.items():
                if query_ngrams[ngram]["lower"] in vis_keywords:
                    return vis_type, query_ngrams[ngram]["lower"]

        return None, None

    def design_has_valid_task(self, designs):
        has_valid_task = False
        # ENSURE if COMBOS has the attributes to which the TASK is applied. If NOT, don"t do anything.
        for task in self.nl4dv_instance.extracted_tasks:
            if task in designs["tasks"]:
                has_valid_task = True

        return has_valid_task

    def get_vis_list(self, attribute_list):
        vis_objects = list()

        # Create combinations of all attributes. Ideally, taking ALL combinations should suffice BUT we also have AMBIGUOUS attributes.
        # Hence, we generate all combinations and then FILTER based on their ambiguity, etc.
        for i in range(1, len(attribute_list) + 1):
            combinations = itertools.combinations(attribute_list, i)
            for combo in combinations:
                if self.nl4dv_instance.attribute_genie_instance.validate_attr_combo(attr_combo=combo, query_phrase=[], allow_subset=False):
                    continue

                # Create a SORTED list of attributes and their datatypes to match the keys of the VisReco dictionary. e.g. `QQ`, `QNO`, ...
                attr_list, attr_type_str = self.nl4dv_instance.attribute_genie_instance.get_attr_datatype_shorthand(combo)

                # The Attribute-Datatype combination (e.g. QQNN) is NOT yet supported. Continue.
                if attr_type_str not in vis_design_combos:
                    continue

                # Is at least one task supported for the Designs. Used to disambiguate/choose between tasks for e.g. `distribution` and `derived value`.
                design_has_valid_task = any([t in vis_design_combos[attr_type_str]["tasks"] for t in self.nl4dv_instance.extracted_tasks])

                # Is at least one vis supported for the Designs. Used to disambiguate/choose between mark types for e.g. `bar` and `tick`.
                design_has_valid_vis = self.nl4dv_instance.extracted_vis_type is not None and self.nl4dv_instance.extracted_vis_type in vis_design_combos[attr_type_str]["visualizations"]

                # NL4DV does not support ALL attribute type combinations yet, e.g. T vs T vs T. We don"t have Vega-Lite encodings for these.
                if attr_type_str in vis_design_combos \
                        and vis_design_combos[attr_type_str]["support"]:

                    # For each combination, there are multiple design solutions, e.g. histogram or strip plot for a "quantitative (Q)" attribute
                    for d_counter in range(len(vis_design_combos[attr_type_str]["designs"])):

                        # Create reference to a design that matches the attribute combination.
                        design = copy.deepcopy(vis_design_combos[attr_type_str]["designs"][d_counter])

                        # Filter the DESIGN based on TASKs
                        if design_has_valid_task and design["task"] not in self.nl4dv_instance.extracted_tasks:
                            continue

                        # Filter the DESIGN based on explicit VISUALIZATIONS
                        if design_has_valid_vis and self.nl4dv_instance.extracted_vis_type != design["vis_type"]:
                            continue

                        # Generate Vega-Lite specification along with it"s relevance score for the attribute and task combination.
                        vl_genie_instance = self.get_vis(design, attr_type_str, attr_list)

                        # This is the Score object that helps prioritize between AMBIGUOUS attributes
                        confidence_obj = dict()
                        for attr in attr_list:
                            confidence_obj[attr] = 0 if "confidence" not in self.nl4dv_instance.extracted_attributes[attr]["meta"] else self.nl4dv_instance.extracted_attributes[attr]["meta"]["confidence"]/100

                        if vl_genie_instance is not None:
                            vis_object = {
                                "score": sum(vl_genie_instance.score_obj.values()) + sum(confidence_obj.values()),
                                "scoreObj": vl_genie_instance.score_obj,
                                "confidenceObj": confidence_obj,
                                "attributes": attr_list,
                                "queryPhrase": self.nl4dv_instance.extracted_vis_token,
                                "visType": self.nl4dv_instance.extracted_vis_type,
                                "tasks": list(self.nl4dv_instance.extracted_tasks.keys()),
                                "inferenceType": 'implicit' if self.nl4dv_instance.extracted_vis_type is None else 'explicit',
                                "vlSpec": vl_genie_instance.vl_spec
                            }
                            if vis_object not in vis_objects and vis_object["score"] > 0:
                                vis_objects.append(vis_object)

                else:
                    vis_object = self.create_datatable_vis(attr_list)
                    if vis_object not in vis_objects and vis_object["score"] > 0:
                        vis_objects.append(vis_object)

        return list(sorted(vis_objects, key=lambda o: o["score"], reverse=True))

    def get_vis(self, design, attr_type_combo, attr_list):

        # CREATE a new Vega-Lite Spec
        vl_genie_instance = VLGenie()

        # MAP the attributes to the DESIGN spec.
        for index, attr in enumerate(attr_list):
            dim = design["priority"][index]  # Dimension: x, y, color, size, tooltip, ...
            agg = design[dim]["agg"]  # Aggregate: sum, mean, ...
            datatype = self.nl4dv_instance.data_genie_instance.data_attribute_map[attr]["dataType"]

            # Update the design with the attribute. It could be referenced later.
            design[dim]["attr"] = attr
            design[dim]["is_defined"] = True

            # Set the default VIS mark type. Note: Can be overridden later.
            vl_genie_instance.set_vis_type(design["vis_type"])

            # Set the encoding Note: Can be overridden later.
            vl_genie_instance.set_encoding(dim, attr, datatype, agg)

            # Set Score
            vl_genie_instance.score_obj["by_attributes"] += self.nl4dv_instance.extracted_attributes[attr]["matchScore"]

        # If an attribute is dual-encoded e.g. x axis as well as count of y axis, the attribute is supposed to be encoded to both channels.
        for encoding in design["mandatory"]:
            if not design[encoding]["is_defined"]:
                attr_reference = design[encoding]["attr_ref"]
                attr = design[attr_reference]["attr"]
                datatype = self.nl4dv_instance.data_genie_instance.data_attribute_map[attr]["dataType"]
                agg = design[encoding]["agg"]
                vl_genie_instance.set_encoding(encoding, attr, datatype, agg)

        # ENSURE if COMBOS has the attributes to which the TASK is applied. If NOT, don"t do anything.
        for task in self.nl4dv_instance.extracted_tasks:
            for task_instance in self.nl4dv_instance.extracted_tasks[task]:

                if task == "filter":
                    # If there is NO Datatype Ambiguity, then apply the Filter Task. Else let it be the way it is.
                    # Datatype ambiguity example: "Content Rating > 5" is NOT possible because Content Rating is a Nominal attribute.
                    if not (task_instance["isValueAmbiguous"] and task_instance["meta"]["value_ambiguity_type"] == "datatype"):
                        vl_genie_instance.set_task(None, task_instance)
                        vl_genie_instance.score_obj["by_task"] += task_instance["matchScore"]

                else:
                    # If a NON-FILTER task has an attribute that is NOT in the combos (means it was ambiguous), then No Need to Apply this FILTER.
                    # E.g. We don't want IMDB Rating > 5 to be applied to a VIS design with Rotten Tomatoes Rating
                    if any([attr not in attr_list for attr in task_instance["attributes"]]):
                        continue

                    if task == "derived_value":
                        # If there is NO Datatype Ambiguity, then apply the Derived Value Task. Else let it be the way it is.
                        # Datatype ambiguity example: "SUM(Genre)" is NOT possible because Genre is a Nominal attribute.
                        if not (task_instance["isValueAmbiguous"] and task_instance["meta"]["value_ambiguity_type"] == "datatype"):
                            if design["vis_type"] in ["histogram", "boxplot"]:
                                return None

                            # Iterate over all encodings and if the corresponding attribute matches that in the task, then UPDATE the "aggregate".
                            for dimension in design["mandatory"]:
                                attr = design[dimension]["attr"]
                                if attr in task_instance["attributes"]:
                                    vl_genie_instance.score_obj["by_task"] += task_instance["matchScore"]

                                    datatype = self.nl4dv_instance.data_genie_instance.data_attribute_map[attr]["dataType"]
                                    new_agg = constants.operator_symbol_mapping[task_instance["operator"]]
                                    vl_genie_instance.set_encoding(dimension, attr, datatype, new_agg)

                    elif task == "distribution":
                        # Increment score by_task
                        vl_genie_instance.score_obj["by_task"] += task_instance["matchScore"]

                    elif task == "correlation":
                        # For correlations, there should be NO aggregation between the attributes
                        for dimension in design['mandatory']:
                            if design[dimension]["attr"] in task_instance["attributes"]:
                                # If there exists some aggregate already, then this is a CONFLICT and we should DEDUCT points
                                if design[dimension]['agg'] is not None:
                                    vl_genie_instance.score_obj["by_task"] -= 1

                                design[dimension]['agg'] = None
                                vl_genie_instance.set_encoding_aggregate(dimension, None)

                        # Correlation < scatterplot (mark type = point)
                        vl_genie_instance.set_vis_type("scatterplot")

                        # Increment score by_task
                        vl_genie_instance.score_obj["by_task"] += task_instance["matchScore"]

                    elif task == "find_extremum":
                        # If there is NO Datatype Ambiguity, then apply the Derived Value Task. Else let it be the way it is.
                        # Datatype ambiguity example: "SUM(Genre)" is NOT possible because Genre is a Nominal attribute.
                        if not (task_instance["isValueAmbiguous"] and task_instance["meta"]["value_ambiguity_type"] == "datatype"):

                            # Iterate over all encodings and if the corresponding attribute matches that in the task, then UPDATE the "aggregate".
                            for dimension in design["mandatory"]:
                                attr = design[dimension]["attr"]
                                if attr in task_instance["attributes"]:
                                    vl_genie_instance.score_obj["by_task"] += task_instance["matchScore"]
                                    vl_genie_instance.set_task(dimension, task_instance)

                    elif task == "trend":
                        pass

        # If explicit VIS is specified, then override it
        # TODO:- There a few vis (mark) types that are NOT sensible, e.g. asking a scatterplot for a piechart design or a linechart for a boxplot base design. Filter these designs out!
        if self.nl4dv_instance.extracted_vis_type:

            # A design with PIECHART / DONUTCHART as a base should NOT be attempted to be transformed for a different mark type. Note: It has thetas, colors as opposed to x, y.
            if self.nl4dv_instance.extracted_vis_type not in ["piechart", "donutchart"] and design["vis_type"] in ["piechart", "donutchart"]:
                return None

            # PIE CHART + DONUT CHART
            # Can happen between 2 attributes {QN, QO} combinations
            if self.nl4dv_instance.extracted_vis_type in ["piechart", "donutchart"]:
                if attr_type_combo not in ["QN", "QO"]:
                    print("Pie Chart not compatible / not supported for your query.")
                    return None

            # HISTOGRAM
            elif self.nl4dv_instance.extracted_vis_type == "histogram":
                if attr_type_combo not in ["Q", "N", "O", "T"]:
                    print("Histogram not compatible / not supported for your query.")
                    return None

            # STRIP PLOT
            elif self.nl4dv_instance.extracted_vis_type == "stripplot":
                # Stripplot is indicative of a DISTRIBUTION Task. All aggregations should be removed.
                for dimension in design['mandatory']:
                    # If there exists some aggregate already, then this is a CONFLICT and we should DEDUCT points
                    if design[dimension]['agg'] is not None:
                        vl_genie_instance.score_obj["by_vis"] -= 1

                    design[dimension]['agg'] = None
                    vl_genie_instance.set_encoding_aggregate(dimension, None)

            # BAR CHART
            elif self.nl4dv_instance.extracted_vis_type == "barchart":
                pass

            # LINE CHART
            elif self.nl4dv_instance.extracted_vis_type == "linechart":
                pass

            # AREA CHART
            elif self.nl4dv_instance.extracted_vis_type == "areachart":
                if design["vis_type"] == "barchart":
                    return None

            # SCATTERPLOT
            elif self.nl4dv_instance.extracted_vis_type == "scatterplot":
                # For scatterplots, treat it as a Correlation task. There should be NO aggregation between the attributes,
                # and mark type should be "point"
                for dimension in design['mandatory']:
                    # If there exists some aggregate already, then this is a CONFLICT and we should DEDUCT points
                    if design[dimension]['agg'] is not None:
                        vl_genie_instance.score_obj["by_task"] -= 1

                    design[dimension]['agg'] = None
                    vl_genie_instance.set_encoding_aggregate(dimension, None)

                    # Correlation < scatterplot (mark type = point)
                    vl_genie_instance.set_vis_type("scatterplot")

            # BOX PLOT
            elif self.nl4dv_instance.extracted_vis_type == "boxplot":
                if "Q" not in attr_type_combo:
                    print("Box Plot requires at least one continuous axis. Not compatible / supported for your query.")
                    return None

            # Set the VIS mark type in the vl_genie_instance
            vl_genie_instance.set_vis_type(self.nl4dv_instance.extracted_vis_type)

            # just here because the user/developer explicitly requested this
            vl_genie_instance.score_obj["by_vis"] += self.nl4dv_instance.match_scores['vis']['explicit']

        else:
            # There are a few designs tagged as "not_suggested_by_default",
            # e.g., in absence of a task, there's no need to show both DERIVED_VALUE (barchart + mean) and DISTRIBUTION (stripplot) implicit tasked visualizations
            if design["not_suggested_by_default"]:
                return None

        # Encode the label attribute as a TOOLTIP to show the dataset label on hover.
        # Note: This will ONLY be added when there is NO aggregation, i.e., all data points are visible.
        vl_genie_instance.add_label_attribute_as_tooltip(self.nl4dv_instance.label_attribute)

        # AESTHETICS
        # ------------------
        # Format ticks (e.g. 10M, 1k, ... ) for Quantitative axes
        vl_genie_instance.add_tick_format()
        # ------------------

        # Enable Tooltips
        # ------------------
        vl_genie_instance.add_tooltip()
        # ------------------

        #  Finally, let"s set the data and Rock"n Roll!
        # ------------------
        vl_genie_instance.set_data(self.nl4dv_instance.data_url)
        # ------------------

        return vl_genie_instance

    # Return a Data Table in Vega-Lite
    def create_datatable_vis(self, sorted_combo):

        # Create a new base Vega-Lite Spec
        vl_genie_instance = VLGenie()

        # Set the explicit_vis_type to a datatable and then make relevant transforms there.
        vl_genie_instance.set_vis_type("datatable")

        for attr in sorted_combo:
            # Create a column with mark type = text
            vl_genie_instance.create_and_add_column_to_datatable(attr)

            # Append the scores
            vl_genie_instance.score_obj["by_attributes"] += self.nl4dv_instance.extracted_attributes[attr]["matchScore"]

        #  Set the data
        vl_genie_instance.set_data(self.nl4dv_instance.data_url)

        # Create the Visualization object to return
        vis_object = {
            "score": sum(vl_genie_instance.score_obj.values()),
            "scoreObj": vl_genie_instance.score_obj,
            "attributes": sorted_combo,
            "visType": "datatable",
            "queryPhrase": None,
            "tasks": list(self.nl4dv_instance.extracted_tasks.keys()),
            "inferenceType": 'implicit' if self.nl4dv_instance.extracted_vis_type is None else 'explicit',
            "vlSpec": vl_genie_instance.vl_spec
        }

        return vis_object




from nl4dv import NL4DV
from nl4dv.utils import helpers
import json
import pytest
import pandas as pd
import os


class TestQueriesClass:

    _files = [
              '1-fullyspecified-attributes-tasks-vis.txt',
              '2-underspecified-attributes-tasks.txt',
              '3-underspecified-attributes-vis.txt',
              '4-underspecified-attributes.txt',
              'debugger_batch_queries.txt',
              'vis_matrix_queries.tsv'
              ]

    # Dependency Parser OPTIONS
    dependency_parser_config = {
        "corenlp": {'name': 'corenlp', 'model': os.path.join("examples","assets","jars","stanford-english-corenlp-2018-10-05-models.jar"),'parser': os.path.join("examples","assets","jars","stanford-parser.jar")},
        "corenlp-server": {'name': 'corenlp-server', 'url': 'http://localhost:9000'},
        "spacy": {'name': 'spacy','model': 'en_core_web_sm'}
    }

    @staticmethod
    @pytest.fixture(scope='class')
    def set_dependency_parser(request):
        # Base NL4DV instance
        return request.param

    @staticmethod
    def init_nl4dv(dataset, alias_file):
        # Base NL4DV instance
        nl4dv_instance = NL4DV(verbose=False)
        nl4dv_instance.set_data(data_url=os.path.join("examples","assets","data",dataset))
        nl4dv_instance.set_alias_map(alias_url=os.path.join("examples","assets","aliases", alias_file))
        return nl4dv_instance

    @pytest.fixture(scope='class')
    def read_queries(self, request):
        query_file = os.path.join('examples', 'assets', 'queries', request.param)
        df = pd.read_csv(query_file, sep='\t')
        return df.where(pd.notnull(df), None)

    @pytest.mark.parametrize("set_dependency_parser", list(dependency_parser_config.keys()), indirect=True, scope="class")
    @pytest.mark.parametrize("read_queries", _files, indirect=True, scope="class")
    def test_queries(self, read_queries, set_dependency_parser):
        nl4dv_instance, dataset, errors = None, None, list()
        for index, row in read_queries.iterrows():
            if dataset != row['Datafile']:
                dataset = row['Datafile']
                alias_file = row['Aliasfile']
                nl4dv_instance = TestQueriesClass.init_nl4dv(dataset, alias_file)

            nl4dv_instance.set_dependency_parser(config=TestQueriesClass.dependency_parser_config[set_dependency_parser])

            if nl4dv_instance is not None:
                output = json.loads(json.dumps(nl4dv_instance.analyze_query(row['Query']), default=helpers.set_default))

                if row["Attributes"] and set(output["attributeMap"].keys()) != set(row["Attributes"].split(",")):
                    print(row['Query'],row['Attributes'],output["attributeMap"].keys())
                    errors.append(1)

                if row["Tasks"] and set(output["taskMap"].keys()) != set(row["Tasks"].split(",")):
                    print(row['Query'],row['Tasks'])
                    errors.append(1)

        assert len(errors) == 0



from setuptools import setup, find_packages
from os import path

# read the contents of the README file
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='nl4dv',
    version='0.0.1',
    author='Arpit Narechania',
    author_email='arpitnarechania@gatech.edu',
    packages=find_packages(),
    include_package_data=True,
    setup_requires=['pytest-runner'],
    test_requires=['pytest', 'pytest-cov'],
    scripts=[],
    url='https://github.com/nl4dv/nl4dv',
    license='LICENSE',
    description='Natural Language Toolkit for Data Visualization',
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=[
        "dateparser~=0.7.6",
        "fuzzywuzzy~=0.8.1",
        "nltk~=3.5",
        "pytest~=3.10.1",
        "pytest-cov~=2.6.0",
        "python-dateutil~=2.6.1",
        "python-Levenshtein~=0.12.0",
        "si-prefix~=1.2.2",
        "spacy~=2.2",
        "vega~=3.1"
    ]
)



# In-built Libraries
import os
import time
from collections import OrderedDict

# Third-Party Libraries
import spacy
from nltk.stem.porter import PorterStemmer
from nltk.parse.stanford import StanfordDependencyParser
from nltk.parse.corenlp import CoreNLPDependencyParser

from vega import VegaLite

# NL4DV Imports
from nl4dv.datagenie import DataGenie
from nl4dv.querygenie import QueryGenie
from nl4dv.attributegenie import AttributeGenie
from nl4dv.taskgenie import TaskGenie
from nl4dv.visgenie import VisGenie
from nl4dv.utils import helpers, constants, error_codes


class NL4DV:
    """
    Class exposed to users to interact with the package. Exposes modules in the package via
    public methods

    """

    def __init__(self, data_url=None,
                 alias_url=None,
                 alias_map=None,
                 label_attribute=None,
                 ignore_words=list(),
                 reserve_words=list(),
                 dependency_parser_config=None,
                 verbose=False):

        # inputs
        self.data_url = data_url
        self.alias_url = alias_url
        self.alias_map = alias_map
        self.label_attribute = label_attribute
        self.ignore_words = ignore_words
        self.reserve_words = reserve_words
        self.verbose = verbose

        # outputs
        self.execution_durations = dict()
        self.query_raw = None
        self.query_processed = ""
        self.query_tokens = list()
        self.query_ngrams = dict()
        self.extracted_vis_type = None
        self.extracted_vis_token = None
        self.extracted_tasks = OrderedDict()
        self.extracted_attributes = OrderedDict()
        self.vis_list = None
        self.dependencies = list()

        # Load constants: thresholds, mappings, scores
        self.vis_keyword_map = constants.vis_keyword_map
        self.task_keyword_map = constants.task_keyword_map
        self.match_scores = constants.match_scores
        self.match_thresholds = constants.match_thresholds

        # Others
        self.dialog = False
        self.debug = False

        # initialize porter stemmer instance
        self.porter_stemmer_instance = PorterStemmer()

        # Set the dependency parser
        self.dependency_parser = None
        self.dependency_parser_instance = None
        self.set_dependency_parser(dependency_parser_config)

        # Internal Class Instances
        self.data_genie_instance = DataGenie(self)  # initialize a DataGenie instance.
        self.query_genie_instance = QueryGenie(self)  # initialize a QueryGenie instance.
        self.attribute_genie_instance = AttributeGenie(self)   # initialize a AttributeGenie instance.
        self.task_genie_instance = TaskGenie(self)  # initialize a TaskGenie instance.
        self.vis_genie_instance = VisGenie(self)   # initialize a VisGenie instance.

    # returns a VegaLite object of the best (1st) visualization after analyzing the query.
    def render_vis(self, query_raw):
        # type: (str) -> VegaLite
        response = self.analyze_query(query_raw)
        if len(response['visList']) == 0:
            print("No best Viz; please try again.")
            return VegaLite({})
        return VegaLite(response['visList'][0]['vlSpec'])

    # ToDo:- Discuss support for non-ascii characters? Fallback from unicode to ascii good enough?
    # ToDo:- Discuss ERROR Handling
    # ToDo:- Utilities to perform unit conversion (eg. seconds > minutes). Problem: Tedious to infer base unit from data. - LATER
    def analyze_query(self, query_raw, dialog=False, debug=False):
        # type: (str) -> dict

        self.execution_durations = dict()
        self.dialog = dialog
        self.debug = debug

        # If not a follow-up query, reset the output variables.
        if not dialog:
            self.extracted_vis_type = None
            self.extracted_vis_token = None
            self.extracted_tasks = OrderedDict()
            self.extracted_attributes = OrderedDict()
            self.vis_list = None

        # CLEAN AND PROCESS QUERY
        self.query_raw = query_raw
        helpers.cond_print("Raw Query: " + self.query_raw, self.verbose)
        st = time.time()
        self.query_processed = self.query_genie_instance.process_query(self.query_raw)
        self.query_tokens = self.query_genie_instance.clean_query_and_get_query_tokens(self.query_processed, self.reserve_words, self.ignore_words)
        self.query_ngrams = self.query_genie_instance.get_query_ngrams(' '.join(self.query_tokens))
        self.dependencies = self.query_genie_instance.create_dependency_tree(self.query_processed)
        helpers.cond_print("Processed Query: " + self.query_processed, self.verbose)
        self.execution_durations['clean_query'] = time.time() - st

        # DETECT EXPLICIT AND IMPLICIT ATTRIBUTES
        st = time.time()
        self.extracted_attributes = self.attribute_genie_instance.extract_attributes(self.query_ngrams)
        helpers.cond_print("Final Extracted Attributes: " + str(list(self.extracted_attributes.keys())), self.verbose)
        self.execution_durations['extract_attributes'] = time.time() - st

        # DETECT EXPLICIT VISUALIZATION UTTERANCES
        st = time.time()
        self.extracted_vis_type, self.extracted_vis_token = self.vis_genie_instance.extract_vis_type(self.query_ngrams)
        self.execution_durations['extract_vis_type'] = time.time() - st

        # DETECT IMPLICIT AND EXPLICIT TASKS
        st = time.time()
        task_map = self.task_genie_instance.extract_explicit_tasks_from_dependencies(self.dependencies)

        # Filters from Domain Values
        task_map = self.task_genie_instance.extract_explicit_tasks_from_domain_value(task_map)

        # At this stage, which attributes are encodeable?
        encodeable_attributes = self.attribute_genie_instance.get_encodeable_attributes()

        # INFER tasks based on (encodeable) attribute Datatypes
        task_map = self.task_genie_instance.extract_implicit_tasks_from_attributes(task_map, encodeable_attributes)

        # From the generated TaskMap, ensure that the task "keys" are NOT EMPTY LISTS
        self.extracted_tasks = self.task_genie_instance.filter_empty_tasks(task_map)
        self.execution_durations['extract_tasks'] = time.time() - st

        # RECOMMEND VISUALIZATIONS FROM ATTRIBUTES, TASKS, and VISUALIZATIONS
        st = time.time()

        # Final list of encodeable attributes in the VIS
        final_encodeable_attributes = self.attribute_genie_instance.update_encodeable_attributes_based_on_tasks()

        self.vis_list = self.vis_genie_instance.get_vis_list(attribute_list=final_encodeable_attributes)
        self.execution_durations['get_vis_list'] = time.time() - st
        self.execution_durations['total'] = sum(self.execution_durations.values())

        # Prepare output
        output = {
            'status': 'SUCCESS' if len(self.vis_list) > 0 else 'FAILURE',
            'debug': {'execution_durations': self.execution_durations},
            'query_raw': self.query_raw,
            'query': self.query_processed,
            'dataset': self.data_url,
            'visList': self.vis_list,
            'attributeMap': self.extracted_attributes,
            'taskMap': self.extracted_tasks,
            'followUpQuery': self.dialog,
            'contextObj': None
        }

        return output if debug else helpers.delete_keys_from_dict(output, keys=constants.keys_to_delete_in_output)

    # Update the attribute datatypes that were not correctly detected by NL4DV
    def set_attribute_datatype(self, attr_type_obj):
        return self.data_genie_instance.set_attribute_datatype(attr_type_obj=attr_type_obj)

    # Set Label attribute for the dataset, i.e. one that defines what the dataset is about.
    # e.g. "Correlate horsepower and MPG for sports car models" should NOT apply an explicit attribute for models since there are two explicit attributes already present.
    def set_label_attribute(self, label_attribute):
        return self.data_genie_instance.set_label_attribute(label_attribute=label_attribute)

    # WORDS that should be IGNORED in the query, i.e. NOT lead to the detection of attributes and tasks
    # `Movie` in movies dataset
    # `Car` in cars dataset
    def set_ignore_words(self, ignore_words):
        return self.data_genie_instance.set_ignore_words(ignore_words=ignore_words)

    # Custom STOPWORDS that should NOT removed from the query, as they might be present in the domain.
    # e.g. `A` in grades dataset
    def set_reserve_words(self, reserve_words):
        return self.data_genie_instance.set_reserve_words(reserve_words=reserve_words)

    # Sets the AliasMap
    def set_alias_map(self, alias_map=None, alias_url=None):
        return self.data_genie_instance.set_alias_map(alias_map=alias_map, alias_url=alias_url)

    # Sets the Dataset
    def set_data(self, data_url=None):
        return self.data_genie_instance.set_data(data_url=data_url)

    # Sets the String Matching, Domain Word Limit, ... Thresholds
    def set_thresholds(self, thresholds):
        for t in thresholds:
            if t in self.match_thresholds and (isinstance(thresholds[t], float) or isinstance(thresholds[t], int)):
                self.match_thresholds[t] = thresholds[t]
        return True

    # Sets the Scoring Weights for the way attributes / tasks and visualizations are detected.
    def set_importance_scores(self, scores):
        for s in scores.keys():
            if s in self.match_scores and isinstance(scores[s], float):
                self.match_scores[s] = scores[s]

        return True

    # Get the dataset metadata
    def get_metadata(self):
        return self.data_genie_instance.data_attribute_map

    # Create a dependency parser instance
    def set_dependency_parser(self, config):
        if isinstance(config, dict):
            helpers.cond_print("Dependency Parser: " + config["name"], self.verbose)
            self.dependency_parser = config["name"]
            if config["name"] == "spacy":
                """
                    Sets the model and returns the Spacy NLP instance. Example ways from the Spacy docs:
                    spacy.load("en") # shortcut link
                    spacy.load("en_core_web_sm") # package
                    spacy.load("/path/to/en") # unicode path
                    spacy.load(Path("/path/to/en")) # pathlib Path
                """
                self.dependency_parser_instance = spacy.load(config["model"])

            elif config["name"] == "corenlp":
                if 'CLASSPATH' not in os.environ:
                    os.environ['CLASSPATH'] = ""

                cpath = config["model"] + os.pathsep + config["parser"]
                if cpath not in os.environ['CLASSPATH']:
                    os.environ['CLASSPATH'] = cpath + os.pathsep + os.environ['CLASSPATH']

                # TODO:- DEPRECATED
                self.dependency_parser_instance = StanfordDependencyParser(path_to_models_jar=config["model"],
                                                                           encoding='utf8')
            elif config["name"] == "corenlp-server":
                # Requires the CoreNLPServer running in the background at the below URL (generally https://localhost:9000)
                # Start server by running the following command in the JARs directory.
                # `java -mx4g -cp "*" edu.stanford.nlp.pipeline.StanfordCoreNLPServer -annotators "tokenize,ssplit,pos,lemma,parse,sentiment" -port 9000 -timeout 30000`
                self.dependency_parser_instance = CoreNLPDependencyParser(url=config["url"])



import csv
from collections import Counter
import json
import sys
from dateparser import parse
from nl4dv.utils import constants, error_codes, helpers
import os

class DataGenie:
    """
    Pre-process data attributes into different categories
    """

    def __init__(self, nl4dv_instance):

        # nl4dv instance
        self.nl4dv_instance = nl4dv_instance
        
        # set label attribute
        self.set_label_attribute(label_attribute=self.nl4dv_instance.label_attribute)

        # set ignore list
        self.set_ignore_words(ignore_words=self.nl4dv_instance.ignore_words)

        # set custom stopwords list
        self.set_reserve_words(reserve_words=self.nl4dv_instance.reserve_words)

        # set data and extract attributes
        self.set_data(data_url=self.nl4dv_instance.data_url)

        # set alias map
        self.set_alias_map(alias_url=self.nl4dv_instance.alias_url, alias_map=self.nl4dv_instance.alias_map)

    # Update the attribute datatypes that were not correctly detected by NL4DV
    def set_attribute_datatype(self, attr_type_obj):
        # Set new datatype
        for attribute, data_type in attr_type_obj.items():
            self.data_attribute_map[attribute]['dataType'] = data_type

            if data_type in constants.attribute_types.values():
                self.populate_dataset_meta_for_attr(attribute, data_type)
            else:
                helpers.cond_print("Invalid Target DataType. Choose from " + str(constants.attribute_types.values()), debug=True)
                sys.exit(error_codes.BAD_INPUT_ATTRIBUTE_DATA_TYPE)

    # Set Label attribute for the dataset, i.e. one that defines what the dataset is about.
    # e.g. "Correlate horsepower and MPG for sports car models" should NOT apply an explicit attribute for models since there are two explicit attributes already present.
    def set_label_attribute(self, label_attribute):
        self.nl4dv_instance.label_attribute = label_attribute
        return True

    # WORDS that should be IGNORED in the query, i.e. NOT lead to the detection of attributes and tasks
    # `Movie` in movies dataset
    # `Car` in cars dataset
    def set_ignore_words(self, ignore_words):
        self.nl4dv_instance.ignore_words = ignore_words
        return True

    # Custom STOPWORDS that should NOT removed from the query, as they might be present in the domain.
    # e.g. `A` in grades dataset
    def set_reserve_words(self, reserve_words):
        self.nl4dv_instance.reserve_words = reserve_words
        return True

    # Sets the Dataset
    def set_data(self, data_url=None):
        # type: (str) -> bool
        """
        User can choose to manually initialize data

        """
        self.nl4dv_instance.data_url = data_url

        # initialize values
        self.data_attribute_map = dict()
        self.data = list()
        self.rows = 0

        if self.nl4dv_instance.data_url is not None and os.path.isfile(self.nl4dv_instance.data_url):

            # local variables
            reader = None
            json_data = None
            attributes = list()

            if self.nl4dv_instance.data_url.lower().endswith('.csv'):
                reader = csv.reader(open(self.nl4dv_instance.data_url, 'r', encoding='utf-8'),delimiter=',')
                attributes = next(reader)  # assumes headers are in the first line
            elif self.nl4dv_instance.data_url.lower().endswith('.tsv'):
                reader = csv.reader(open(self.nl4dv_instance.data_url, 'r', encoding='utf-8'),delimiter='\t')
                attributes = next(reader)  # assumes headers are in the first line
            elif self.nl4dv_instance.data_url.lower().endswith('.json'):
                json_data = json.load(open(self.nl4dv_instance.data_url, 'r', encoding='utf-8'))
                attributes = json_data[0].keys()

            # initialize properties in Attribute Map
            for attr in attributes:
                self.data_attribute_map[attr] = {
                    'domain': set(),
                    'isLabelAttribute': attr == self.nl4dv_instance.label_attribute,
                    'summary': dict(),
                    'dataTypeList': list(), # temporary to determine datatype
                    'dataType': '',
                    'aliases': list(),
                }

            # initialize properties in Attribute Map
            # implies file is either .csv or .tsv
            if reader is not None:
                for line in reader:
                    data_obj = dict()
                    for i in range(len(line)):
                        data_obj[attributes[i]] = line[i]
                    self.data.append(data_obj)
                    self.rows += 1
            else:
                # JSON file
                for data_obj in json_data:
                    self.data.append(data_obj)
                    self.rows += 1

            # infer attribute datatypes and compute summary (range, domain)
            for datum in self.data:
                for attr in self.data_attribute_map.keys():
                    attr_val = datum[attr]

                    # Check for Numeric (float, int)
                    if helpers.isfloat(attr_val) or helpers.isint(attr_val):
                        attr_datatype = constants.attribute_types['QUANTITATIVE']
                        self.populate_dataset_meta(attr, attr_val, attr_datatype)

                    # Check for Datetime
                    # ToDo:- Works fine for datetime strings. Not for others like Epochs and Int-only Years (e.g. 2018) which get captured above.
                    # ToDo:- It is VERY risky to switch this elif block with the if block above
                    elif helpers.isdate(attr_val)[0]:
                        attr_datatype = constants.attribute_types['TEMPORAL']
                        self.populate_dataset_meta(attr, attr_val, attr_datatype)

                    # Otherwise set as Nominal
                    else:
                        attr_datatype = constants.attribute_types['NOMINAL']
                        self.populate_dataset_meta(attr, attr_val, attr_datatype)

                    # Irrespective of above assignment, make a list of attribute types for each data row
                    # to take best decision on heterogeneous data with multiple datatypes
                    self.data_attribute_map[attr]['dataTypeList'].append(attr_datatype)

            # Determine the Datatype based on majority of values.
            # Also Override a few datatypes set above based on rules such as NOMINAL to ORDINAL if all values are unique such as Sr. 1, Sr. 2, ...
            for attr in self.data_attribute_map:
                # most common attribute type
                attr_datatype = Counter(self.data_attribute_map[attr]['dataTypeList']).most_common(1)[0][0]

                # if it's quantitative but with less than or equal to 12 unique values, then it's ordinal.
                # eg. 1, 2, 3, ..., 12 (months of a year)
                # eg. -3, -2, -1, 0, 1, 2, 3 (likert ratings)
                if attr_datatype == constants.attribute_types['QUANTITATIVE'] and len(
                        self.data_attribute_map[attr]['domain']) <= 12:
                    attr_datatype = constants.attribute_types['ORDINAL']
                    self.populate_dataset_meta_for_attr(attr, attr_datatype)

                # If an attribute has (almnost) no repeating value, then mark it as the label attribute.
                # eg. primary/unique key of the table? Car1 , Car2, Car3, ...
                # Almost == 90% heuristic-based
                if attr_datatype == constants.attribute_types['NOMINAL'] and len(self.data_attribute_map[attr]['domain']) > 0.9 * self.rows:
                    self.nl4dv_instance.label_attribute = attr
                    self.data_attribute_map[attr]['isLabelAttribute'] = True

                # Set the final data type
                self.data_attribute_map[attr]['dataType'] = attr_datatype

                # Presentation
                self.prepare_output(attr, attr_datatype)


            return True

        return False

    # Sets the Alias Map
    def set_alias_map(self, alias_map=None, alias_url=None):
        # type: (dict, str) -> bool
        """
        User can choose to manually initialize data

        """
        self.nl4dv_instance.alias_url= alias_url
        self.nl4dv_instance.alias_map = alias_map

        if self.nl4dv_instance.alias_url is not None and os.path.isfile(self.nl4dv_instance.alias_url):
            self.nl4dv_instance.alias_map = json.load(open(self.nl4dv_instance.alias_url, 'r', encoding='utf-8'))

        if self.nl4dv_instance.alias_map is not None:
            for attr in self.nl4dv_instance.alias_map:
                if attr in self.data_attribute_map:
                    self.data_attribute_map[attr]['aliases'].extend(self.nl4dv_instance.alias_map[attr])

        return True

    # Set Domain, Ranges for Attributes of different datatypes
    def populate_dataset_meta(self, attr, attr_val, attr_datatype):
        if attr_datatype == constants.attribute_types['QUANTITATIVE']:
            try:
                attr_val = float(attr_val)
            except Exception as e:
                attr_val = float('NaN')
            self.data_attribute_map[attr]['domain'].add(attr_val)

            # Compute Max and Min of the attribute values
            if 'min' not in self.data_attribute_map[attr]['summary']:
                self.data_attribute_map[attr]['summary']['min'] = float("inf")
            if 'max' not in self.data_attribute_map[attr]['summary']:
                self.data_attribute_map[attr]['summary']['max'] = float("-inf")

            if attr_val > self.data_attribute_map[attr]['summary']['max']:
                self.data_attribute_map[attr]['summary']['max'] = attr_val
            if attr_val < self.data_attribute_map[attr]['summary']['min']:
                self.data_attribute_map[attr]['summary']['min'] = attr_val

        elif attr_datatype == constants.attribute_types['TEMPORAL']:
            parsed_status, parsed_attr_val = helpers.isdate(attr_val)
            if parsed_status:
                self.data_attribute_map[attr]['domain'].add(parsed_attr_val)
            else:
                parsed_attr_val = float('NaN')

            # Compute Max and Min of the attribute datetime
            if 'start' not in self.data_attribute_map[attr]['summary']:
                self.data_attribute_map[attr]['summary']['start'] = parsed_attr_val

            if 'end' not in self.data_attribute_map[attr]['summary']:
                self.data_attribute_map[attr]['summary']['end'] = parsed_attr_val

            # print(parsed_status, attr_val, parsed_attr_val, self.data_attribute_map[attr]['summary']['end'])
            if parsed_attr_val > self.data_attribute_map[attr]['summary']['end']:
                self.data_attribute_map[attr]['summary']['end'] = parsed_attr_val

            if parsed_attr_val < self.data_attribute_map[attr]['summary']['start']:
                self.data_attribute_map[attr]['summary']['start'] = parsed_attr_val

        else:
            attr_val = str(attr_val)
            self.data_attribute_map[attr]['domain'].add(attr_val)
            # Compute # occurrences of attribute values
            if 'group_counts' not in self.data_attribute_map[attr]['summary']:
                self.data_attribute_map[attr]['summary']['group_counts'] = dict()
            if attr_val not in self.data_attribute_map[attr]['summary']['group_counts']:
                self.data_attribute_map[attr]['summary']['group_counts'][attr_val] = 0
            self.data_attribute_map[attr]['summary']['group_counts'][attr_val] += 1

    # Set Domain, Ranges for single attribute
    def populate_dataset_meta_for_attr(self, attr, attr_datatype):
        # Re-initialize the dataTypeList to re-set the domain based on the new Prototype
        self.data_attribute_map[attr]['dataTypeList'] = list()

        # For each value in the dataset, add to the attribute map (domain, range)
        self.data_attribute_map[attr]['domain'] = set()
        for datum in self.data:
            attr_val = datum[attr]
            self.populate_dataset_meta(attr, attr_val, attr_datatype)

        # Presentation
        self.prepare_output(attr, attr_datatype)

    def prepare_output(self, attr, attr_datatype):
        # Remove NANs which are supposedly NOT unique
        self.data_attribute_map[attr]['domain'] = set(filter(lambda x: x == x, self.data_attribute_map[attr]['domain']))

        # Convert the set into a sorted list
        try:
            # works for all numbers, all strings
            if attr_datatype in [constants.attribute_types['QUANTITATIVE'], constants.attribute_types['ORDINAL']]:
                self.data_attribute_map[attr]['domain'] = sorted([float(a) for a in self.data_attribute_map[attr]['domain']])
            else:
                self.data_attribute_map[attr]['domain'] = sorted([a for a in self.data_attribute_map[attr]['domain']])
        except Exception as e:
            # works for hybrid numbers + strings combinations
            self.data_attribute_map[attr]['domain'] = sorted([str(a) for a in self.data_attribute_map[attr]['domain']])

        # Delete the keys that are now redundant
        self.delete_unwanted_keys(attr, attr_datatype)

    # Generic method to delete keys based on datatype
    def delete_unwanted_keys(self, attr, attr_datatype):
        # Delete the list of datatypes, not needed anymore
        if 'dataTypeList' in self.data_attribute_map[attr]:
            del self.data_attribute_map[attr]['dataTypeList']

        if attr_datatype == constants.attribute_types['NOMINAL'] or attr_datatype == constants.attribute_types['ORDINAL']:
            self.del_temporal_keys(attr)
            self.del_quantitative_keys(attr)

        elif attr_datatype == constants.attribute_types['QUANTITATIVE']:
            self.del_temporal_keys(attr)
            self.del_nominal_keys(attr)

        elif attr_datatype == constants.attribute_types['TEMPORAL']:
            self.del_quantitative_keys(attr)
            self.del_nominal_keys(attr)

    # Delete Keys from attribute that is determined to be Temporal
    def del_temporal_keys(self, attr):
        if 'start' in self.data_attribute_map[attr]['summary']:
            del self.data_attribute_map[attr]['summary']['start']
        if 'end' in self.data_attribute_map[attr]['summary']:
            del self.data_attribute_map[attr]['summary']['end']

    # Delete Keys from attribute that is determined to be Quantitative
    def del_quantitative_keys(self, attr):
        # Keys from Quantitative
        if 'min' in self.data_attribute_map[attr]['summary']:
            del self.data_attribute_map[attr]['summary']['min']
        if 'max' in self.data_attribute_map[attr]['summary']:
            del self.data_attribute_map[attr]['summary']['max']

    # Delete Keys from attribute that is determined to be Nominal
    def del_nominal_keys(self, attr):
        # Keys from Nominal
        if 'group_counts' in self.data_attribute_map[attr]['summary']:
            del self.data_attribute_map[attr]['summary']['group_counts']



# ERRORS
# Series to be followed = 7XX
BAD_INPUT_DATA_URL = 701
BAD_INPUT_DATA_ALIAS_URL = 702
BAD_INPUT_DATA_ALIAS_MAP = 703
BAD_INPUT_ATTRIBUTE_DATA_TYPE = 704
BAD_INPUT_BAD_DATA_LABEL_ATTRIBUTE = 705
BAD_INPUT_DEPENDENCY_PARSER_NOT_SPECIFIED = 706



from nl4dv.utils import constants


class VLGenie():

    def __init__(self):

        self.vl_spec = dict()
        self.vl_spec['$schema'] = 'https://vega.github.io/schema/vega-lite/v4.json'
        self.vl_spec['mark'] = dict()
        self.vl_spec['encoding'] = dict()
        self.vl_spec['transform'] = list()
        self.bin = False

        # Score object
        self.score_obj = {
            "by_attributes": 0,
            "by_task": 0,
            "by_vis": 0
        }

    def set_vis_type(self, vis):

        if vis == 'histogram':
            self.vl_spec['mark']['type'] = 'bar'
            self.bin = True

        elif vis == 'barchart':
            self.vl_spec['mark']['type'] = 'bar'

        elif vis == 'linechart':
            self.vl_spec['mark']['type'] = 'line'

        elif vis == 'areachart':
            self.vl_spec['mark']['type'] = 'area'

        elif vis == 'scatterplot':
            self.vl_spec['mark']['type'] = 'point'

        elif vis == 'boxplot':
            self.vl_spec['mark']['type'] = 'boxplot'

        elif vis == 'stripplot':
            self.vl_spec['mark']['type'] = 'tick'

        elif vis == 'piechart':
            self.vl_spec['mark']['type'] = 'arc'

        elif vis == 'donutchart':
            self.vl_spec['mark']['type'] = 'arc'
            self.vl_spec['mark']['innerRadius'] = 50

        elif vis == 'datatable':
            # Remove unneeded encodings
            del self.vl_spec['mark']
            del self.vl_spec['encoding']

            # Derive a new "row_number" variable with a sequence of numbers (used as index / counter later on)
            self.vl_spec["transform"] = [{
                "window": [{"op": "row_number", "as": "row_number"}]
            }]

            # Horizontally concatenate each attribute's column (VIS with mark type = text)
            self.vl_spec["hconcat"] = []

    def create_and_add_column_to_datatable(self, attr):

            # VL specification for a column of text. Use the row_number as the y_axis encoding to render vertically. Sly!
            column = {
                "width": 150,
                "title": attr,
                "mark": "text",
                "encoding": {
                    "text": {"field": attr, "type": "nominal"},
                    "y": {"field": "row_number", "type": "ordinal", "axis": None}
                }
            }
            self.vl_spec["hconcat"].append(column)

    def unset_encoding(self, dim):
        if dim in self.vl_spec['encoding']:
            del self.vl_spec['encoding'][dim]

    def get_encoding(self, dim):
        return self.vl_spec['encoding'][dim]

    def set_encoding_aggregate(self, dim, aggregate):
        self.vl_spec['encoding'][dim]['aggregate'] = aggregate

    def set_encoding(self, dim, attr, attr_type, aggregate=None):

        self.vl_spec['encoding'][dim] = dict()
        self.vl_spec['encoding'][dim]['field'] = attr
        self.vl_spec['encoding'][dim]['type'] = constants.vl_attribute_types[attr_type]
        self.vl_spec['encoding'][dim]['aggregate'] = aggregate

        if dim == 'x':
            if self.bin:
                self.vl_spec['encoding'][dim]['bin'] = True

    def set_task(self, dim, task):

        if task["task"] == 'find_extremum':
            if dim is None:
                dim = 'y'
            if task["operator"] == 'MIN':
                if dim == 'x':
                    self.vl_spec['encoding']['y']['sort'] = 'x'
                elif dim == 'y':
                    self.vl_spec['encoding']['x']['sort'] = 'y'
            elif task["operator"] == 'MAX':
                if dim == 'x':
                    self.vl_spec['encoding']['y']['sort'] = '-x'
                elif dim == 'y':
                    self.vl_spec['encoding']['x']['sort'] = '-y'

        elif task["task"] == 'filter':
            if task["operator"] == 'IN':
                for attr in task['attributes']:
                    self.vl_spec['transform'].append({'filter': {"field": attr, "oneOf": task["values"]}})
            elif task["operator"] == 'RANGE':
                for attr in task['attributes']:
                    self.vl_spec['transform'].append({"filter": {"field": attr, "range": task["values"]}})
            elif task["operator"] == 'NOT RANGE':
                for attr in task['attributes']:
                    # self.vl_spec['transform'].append({"filter": {"field": attr, "gte": task["values"][1], "lte": task["values"][0]}})
                    self.vl_spec['transform'].append({"filter": {"not": {"field": attr, "range": task["values"]}}})
            else:
                for attr in task['attributes']:
                    symbol = constants.operator_symbol_mapping[task["operator"]]
                    self.vl_spec['transform'].append({'filter':'lower(datum["{}"]) {} {}'.format(attr, symbol, task["values"][0])})

    def set_data(self, dataUrl):
        # type: (list) -> None
        """
        Set domain data for the visualization

        """
        self.vl_spec['data'] = {'url': dataUrl, 'format': {'type': 'csv'}}

    def add_tick_format(self):
        for dim in self.vl_spec['encoding']:
            if dim in ['x','y'] and self.vl_spec['encoding'][dim]['type'] == 'quantitative':
                if 'axis' not in self.vl_spec['encoding'][dim]:
                    self.vl_spec['encoding'][dim]['axis'] = dict()
                self.vl_spec['encoding'][dim]['axis']["format"] = "s"

    def add_tooltip(self):
        self.vl_spec['mark']['tooltip'] = True

    def add_label_attribute_as_tooltip(self, label_attribute):
        # Check if any of the AXES (encodings) have existing aggregations. If not, then add tooltip which is the label
        has_aggregate = False
        for encoding in self.vl_spec['encoding']:
            if 'aggregate' in self.vl_spec['encoding'][encoding] and self.vl_spec['encoding'][encoding]['aggregate'] is not None:
                has_aggregate = True
                break

        if not has_aggregate:
            self.vl_spec['encoding']['tooltip'] = dict()
            self.vl_spec['encoding']['tooltip']["field"] = label_attribute
